__author__ = 'Brian Wickman'
__version__ = '0.8.3'
__license__ = 'MIT'


import sys

from .base import Environment
from .basic import Boolean, Enum, Float, Integer, String
from .choice import Choice
from .composite import Default, Empty, Required, Struct
from .container import List, Map
from .naming import Namable, Ref
from .parsing import MustacheParser
from .typing import Type, TypeCheck, TypeFactory

if sys.version_info < (2, 6, 5):
  raise ImportError("pystachio requires Python >= 2.6.5")
